@@ -28,10 +28,10 @@
 //
 // ============================================================================================
 
-#include "vbe.h"
-#include "vbetables.h"
 
 // defines available
+// enable LFB support (depends upon bochs-vbe-lfb patch)
+#define VBE_HAVE_LFB
 
 // disable VESA/VBE2 check in vbe info
 //#define VBE2_NO_VESA_CHECK
@@ -39,8 +39,10 @@
 // dynamicly generate a mode_info list
 //#define DYN_LIST
 
-// enable unsupported modi in the mode_info list (ie >8bpp)
-//#define LIST_UNSUPPORTED_MODI
+
+#include "vbe.h"
+#include "vbetables.h"
+
 
 // The current OEM Software Revision of this VBE Bios
 #define VBE_OEM_SOFTWARE_REV 0x0002;
@@ -59,10 +61,10 @@ extern char vbebios_product_revision;
 extern Bit16u vbebios_mode_list;
 #endif
 
-#asm
+ASM_START
 // FIXME: 'merge' these (c) etc strings with the vgabios.c strings?
 _vbebios_copyright:
-.ascii       "Bochs/Plex86 VBE(C) 2002 Jeroen Janssen <japj@darius.demon.nl>"
+.ascii       "Bochs/Plex86 VBE(C) 2003 Jeroen Janssen <japj@darius.demon.nl>"
 .byte        0x00
 
 _vbebios_vendor_name:
@@ -74,15 +76,11 @@ _vbebios_product_name:
 .byte        0x00
 
 _vbebios_product_revision:
-.ascii       "$Id: vbe.c,v 1.12 2002/03/10 16:42:59 japj Exp $"
+.ascii       "$Id: vbe.c,v 1.35 2003/11/03 20:57:01 vruppert Exp $"
 .byte        0x00
 
 _vbebios_info_string:
-.ascii      "Bochs VBE Display Adapter"
-.byte	0x0a,0x0d
-.ascii      "(C) 2002 Jeroen Janssen <japj-vbebios@darius.demon.nl>"
-.byte	0x0a,0x0d
-.ascii	"This VBE Bios is released under the GNU LGPL"
+.ascii      "Bochs VBE Display Adapter enabled"
 .byte	0x0a,0x0d
 .byte	0x0a,0x0d
 .byte	0x00
@@ -90,6 +88,7 @@ _vbebios_info_string:
 _no_vbebios_info_string:
 .ascii      "NO Bochs VBE Support available!"
 .byte	0x0a,0x0d
+.byte	0x0a,0x0d
 .byte 0x00
 
 
@@ -98,29 +97,33 @@ _no_vbebios_info_string:
 //        at least until dynamic list creation is working
 _vbebios_mode_list:
 
-#ifdef LIST_UNSUPPORTED_MODI
-.word VBE_VESA_MODE_640X480X565
-.word VBE_VESA_MODE_800X600X565
-.word VBE_VESA_MODE_640X480X888
-.word VBE_VESA_MODE_800X600X888
-.word VBE_OWN_MODE_800X600X8888
-.word VBE_OWN_MODE_1024X768X8888
-#endif
-
-.word VBE_OWN_MODE_320X200X8
 .word VBE_VESA_MODE_640X400X8
 .word VBE_VESA_MODE_640X480X8
+.word VBE_VESA_MODE_800X600X4
 .word VBE_VESA_MODE_800X600X8
 .word VBE_VESA_MODE_1024X768X8
+.word VBE_VESA_MODE_640X480X1555
+.word VBE_VESA_MODE_640X480X565
+.word VBE_VESA_MODE_640X480X888
+.word VBE_VESA_MODE_800X600X1555
+.word VBE_VESA_MODE_800X600X565
+.word VBE_VESA_MODE_800X600X888
+.word VBE_VESA_MODE_1024X768X1555
+.word VBE_VESA_MODE_1024X768X565
+.word VBE_VESA_MODE_1024X768X888
+.word VBE_OWN_MODE_640X480X8888
+.word VBE_OWN_MODE_800X600X8888
+.word VBE_OWN_MODE_1024X768X8888
+.word VBE_OWN_MODE_320X200X8
 .word VBE_VESA_MODE_END_OF_LIST
 #endif
 
-#endasm
+ASM_END
 
 // from rombios.c
-#define PANIC_PORT 0x400
+#define PANIC_PORT 0x501
 
-#asm
+ASM_START
 MACRO HALT
   ;; the HALT macro is called with the line number of the HALT call.
   ;; The line number is then sent to the PANIC_PORT, causing Bochs to
@@ -131,7 +134,7 @@ MACRO HALT
   mov ax,#?1
   out dx,ax 
 MEND
-#endasm
+ASM_END
 
 // DISPI ioport functions
 // FIXME: what if no VBE host side code?
@@ -169,6 +172,18 @@ static void dispi_set_bpp(bpp)
   outw(VBE_DISPI_IOPORT_DATA,bpp);
 }
 
+static Bit16u dispi_get_bpp()
+{
+  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_BPP);
+  return inw(VBE_DISPI_IOPORT_DATA);
+}
+
+static Bit16u dispi_get_enable()
+{
+  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_ENABLE);
+  return inw(VBE_DISPI_IOPORT_DATA);  
+}
+
 void dispi_set_enable(enable)
   Bit16u enable;
 {
@@ -183,9 +198,97 @@ static void dispi_set_bank(bank)
   outw(VBE_DISPI_IOPORT_DATA,bank);
 }
 
+static Bit16u dispi_get_bank()
+{
+  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_BANK);
+  return inw(VBE_DISPI_IOPORT_DATA);
+}
+
+static void dispi_set_bank_farcall()
+{
+ASM_START
+  cmp bx,#0x0100
+  je dispi_set_bank_farcall_get
+  or bx,bx
+  jnz dispi_set_bank_farcall_error
+  push dx
+  mov ax,# VBE_DISPI_INDEX_BANK
+  mov dx,# VBE_DISPI_IOPORT_INDEX
+  out dx,ax
+  pop ax
+  mov dx,# VBE_DISPI_IOPORT_DATA
+  out dx,ax
+  retf
+dispi_set_bank_farcall_get:
+  mov ax,# VBE_DISPI_INDEX_BANK
+  mov dx,# VBE_DISPI_IOPORT_INDEX
+  out dx,ax
+  mov dx,# VBE_DISPI_IOPORT_DATA
+  in ax,dx
+  mov dx,ax
+  retf
+dispi_set_bank_farcall_error:
+  mov ax,#0x014F
+  retf
+ASM_END
+}
+
+static void dispi_set_x_offset(offset)
+  Bit16u offset;
+{
+  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_X_OFFSET);
+  outw(VBE_DISPI_IOPORT_DATA,offset);
+}
+
+static Bit16u dispi_get_x_offset()
+{
+  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_X_OFFSET);
+  return inw(VBE_DISPI_IOPORT_DATA);
+}
+
+static void dispi_set_y_offset(offset)
+  Bit16u offset;
+{
+  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_Y_OFFSET);
+  outw(VBE_DISPI_IOPORT_DATA,offset);
+}
+
+static Bit16u dispi_get_y_offset()
+{
+  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_Y_OFFSET);
+  return inw(VBE_DISPI_IOPORT_DATA);
+}
+
+static void dispi_set_virt_width(width)
+  Bit16u width;
+{
+  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_VIRT_WIDTH);
+  outw(VBE_DISPI_IOPORT_DATA,width);
+}
+
+static Bit16u dispi_get_virt_width()
+{
+  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_VIRT_WIDTH);
+  return inw(VBE_DISPI_IOPORT_DATA);
+}
+/*
+static void dispi_set_virt_height(height)
+  Bit16u height;
+{
+  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_VIRT_HEIGHT);
+  outw(VBE_DISPI_IOPORT_DATA,height);
+}
+*/
+static Bit16u dispi_get_virt_height()
+{
+  outw(VBE_DISPI_IOPORT_INDEX,VBE_DISPI_INDEX_VIRT_HEIGHT);
+  return inw(VBE_DISPI_IOPORT_DATA);
+}
+
+
 // ModeInfo helper function
-static ModeInfoListItem* mode_info_find_mode(mode)
-  Bit16u mode;
+static ModeInfoListItem* mode_info_find_mode(mode, using_lfb)
+  Bit16u mode; Boolean using_lfb;
 {
   ModeInfoListItem  *cur_info=&mode_info_list;
 
@@ -193,12 +296,25 @@ static ModeInfoListItem* mode_info_find_mode(mode)
   {
     if (cur_info->mode == mode)
     {
-      return cur_info;
+      if (!using_lfb)
+      {
+        return cur_info;
+      }
+      else if (cur_info->info.ModeAttributes & VBE_MODE_ATTRIBUTE_LINEAR_FRAME_BUFFER_MODE)
+      {
+        return cur_info;
+      }
+      else
+      {
+        cur_info++;
+      }
+    }
+    else
+    {
+      cur_info++;
     }
-    
-    cur_info++;
   }
-  
+
   return 0;
 }
 
@@ -207,9 +323,9 @@ static ModeInfoListItem* mode_info_find_mode(mode)
  */
 Boolean vbe_has_vbe_display()
 {
-  dispi_set_id(VBE_DISPI_ID0);
+  dispi_set_id(VBE_DISPI_ID2);
 
-  return (dispi_get_id()==VBE_DISPI_ID0);
+  return (dispi_get_id()==VBE_DISPI_ID2);
 }
 
 /** VBE Init - Initialise the Vesa Bios Extension Code
@@ -228,13 +344,13 @@ void vbe_init()
   if (dispi_id!=VBE_DISPI_ID0)
   {
 //FIXME this results in a 'rombios.c' line panic, but it's actually a 'vbe.c' panic
-#asm    
+ASM_START    
     HALT(__LINE__)
-#endasm    
+ASM_END    
   }
-#ifdef DEBUG
-  printf("VBE Bios $Id: vbe.c,v 1.12 2002/03/10 16:42:59 japj Exp $\n");
-#endif  
+//#ifdef DEBUG
+  printf("VBE Bios $Id: vbe.c,v 1.35 2003/11/03 20:57:01 vruppert Exp $\n");
+//#endif  
 }
 
 /** VBE Display Info - Display information on screen about the VBE
@@ -244,21 +360,21 @@ void vbe_display_info()
   // Check for VBE display extension in Bochs
   if (vbe_has_vbe_display())
   {
-    #asm
+    ASM_START
      mov ax,#0xc000
      mov ds,ax
      mov si,#_vbebios_info_string
      call _display_string
-    #endasm  
+    ASM_END  
   }
   else
   {
-    #asm
+    ASM_START
      mov ax,#0xc000
      mov ds,ax
      mov si,#_no_vbebios_info_string
      call _display_string
-    #endasm  
+    ASM_END  
     
   }
 }  
@@ -281,6 +397,7 @@ Bit16u *AX;Bit16u ES;Bit16u DI;
         VbeInfoBlock      vbe_info_block;
         Bit16u            status;
         Bit16u            result;
+        Bit16u            vbe2_info;
 #ifdef DYN_LIST
         Bit16u            *video_mode_list;
 #endif
@@ -292,11 +409,13 @@ Bit16u *AX;Bit16u ES;Bit16u DI;
 #ifdef DEBUG
         printf("VBE vbe_biosfn_return_vbe_info ES%x DI%x AX%x\n",ES,DI,status);
 #endif
+
+        vbe2_info = 0;
+#ifdef VBE2_NO_VESA_CHECK
+#else
         // get vbe_info_block into local variable
         memcpyb(ss, &vbe_info_block, ES, DI, sizeof(vbe_info_block));
 
-#ifdef VBE2_NO_VESA_CHECK
-#else
         // check for VBE2 signature
         if (((vbe_info_block.VbeSignature[0] == 'V') &&
              (vbe_info_block.VbeSignature[1] == 'B') &&
@@ -308,66 +427,70 @@ Bit16u *AX;Bit16u ES;Bit16u DI;
              (vbe_info_block.VbeSignature[2] == 'S') &&
              (vbe_info_block.VbeSignature[3] == 'A')) )
         {
-#endif
-                
+                vbe2_info = 1;
 #ifdef DEBUG
                 printf("VBE correct VESA/VBE2 signature found\n");
 #endif
-                // VBE Signature
-                vbe_info_block.VbeSignature[0] = 'V';
-                vbe_info_block.VbeSignature[1] = 'E';
-                vbe_info_block.VbeSignature[2] = 'S';
-                vbe_info_block.VbeSignature[3] = 'A';
+        }
+#endif
                 
-                // VBE Version supported
-                vbe_info_block.VbeVersion = 0x0200;
-                
-                // OEM String
-                vbe_info_block.OemStringPtr_Seg = 0xc000;
-                vbe_info_block.OemStringPtr_Off = &vbebios_copyright;
-                
-                // Capabilities
-                vbe_info_block.Capabilities[0] = 0;
-                vbe_info_block.Capabilities[1] = 0;
-                vbe_info_block.Capabilities[2] = 0;
-                vbe_info_block.Capabilities[3] = 0;
+        // VBE Signature
+        vbe_info_block.VbeSignature[0] = 'V';
+        vbe_info_block.VbeSignature[1] = 'E';
+        vbe_info_block.VbeSignature[2] = 'S';
+        vbe_info_block.VbeSignature[3] = 'A';
+        
+        // VBE Version supported
+        vbe_info_block.VbeVersion = 0x0200;
+        
+        // OEM String
+        vbe_info_block.OemStringPtr_Seg = 0xc000;
+        vbe_info_block.OemStringPtr_Off = &vbebios_copyright;
+        
+        // Capabilities
+        vbe_info_block.Capabilities[0] = 0;
+        vbe_info_block.Capabilities[1] = 0;
+        vbe_info_block.Capabilities[2] = 0;
+        vbe_info_block.Capabilities[3] = 0;
 
 #ifdef DYN_LIST
-                // FIXME: This doesn't work correctly somehow?
-                // VBE Video Mode Pointer (dynamicly generated from the mode_info_list)
-                vbe_info_block.VideoModePtr_Seg= ES ;//0xc000;
-                vbe_info_block.VideoModePtr_Off= DI + 34;//&(VBEInfoData->Reserved);//&vbebios_mode_list;
+        // FIXME: This doesn't work correctly somehow?
+        // VBE Video Mode Pointer (dynamicly generated from the mode_info_list)
+        vbe_info_block.VideoModePtr_Seg= ES ;//0xc000;
+        vbe_info_block.VideoModePtr_Off= DI + 34;//&(VBEInfoData->Reserved);//&vbebios_mode_list;
 #else
-                // VBE Video Mode Pointer (staticly in rom)
-                vbe_info_block.VideoModePtr_Seg = 0xc000;
-                vbe_info_block.VideoModePtr_Off = &vbebios_mode_list;
+        // VBE Video Mode Pointer (staticly in rom)
+        vbe_info_block.VideoModePtr_Seg = 0xc000;
+        vbe_info_block.VideoModePtr_Off = &vbebios_mode_list;
 
 #endif
 
 #ifdef DYN_LIST
 
-//                video_mode_list=(Bit16u*)&(vbe_info_block.Reserved);
+//      video_mode_list=(Bit16u*)&(vbe_info_block.Reserved);
 
-                do
-                {
-        #ifdef DEBUG
-                        printf("VBE found mode %x => %x\n", cur_info->mode,cur_mode);
-        #endif
-//                        *video_mode_list=cur_info->mode;
-                        vbe_info_block.Reserved[cur_mode] = cur_info->mode;
-                        
-                        cur_info++;
-                        //video_mode_list++;
-                        cur_mode++;
-                } while (cur_info->mode != VBE_VESA_MODE_END_OF_LIST);
+        do
+        {
+#ifdef DEBUG
+                printf("VBE found mode %x => %x\n", cur_info->mode,cur_mode);
+#endif
+//              *video_mode_list=cur_info->mode;
+                vbe_info_block.Reserved[cur_mode] = cur_info->mode;
                 
-                // Add vesa mode list terminator
-                vbe_info_block.Reserved[cur_mode] = VBE_VESA_MODE_END_OF_LIST;
+                cur_info++;
+                //video_mode_list++;
+                cur_mode++;
+        } while (cur_info->mode != VBE_VESA_MODE_END_OF_LIST);
+        
+        // Add vesa mode list terminator
+        vbe_info_block.Reserved[cur_mode] = VBE_VESA_MODE_END_OF_LIST;
 #endif
 
-                // VBE Total Memory (in 64b blocks)
-                vbe_info_block.TotalMemory = VBE_TOTAL_VIDEO_MEMORY_DIV_64K;
+        // VBE Total Memory (in 64b blocks)
+        vbe_info_block.TotalMemory = VBE_TOTAL_VIDEO_MEMORY_DIV_64K;
 
+        if (vbe2_info)
+	{
                 // OEM Stuff
                 vbe_info_block.OemSoftwareRev = VBE_OEM_SOFTWARE_REV;
                 vbe_info_block.OemVendorNamePtr_Seg = 0xc000;
@@ -379,20 +502,15 @@ Bit16u *AX;Bit16u ES;Bit16u DI;
 
                 // copy updates in vbe_info_block back
                 memcpyb(ES, DI, ss, &vbe_info_block, sizeof(vbe_info_block));
+        }
+	else
+	{
+                // copy updates in vbe_info_block back (VBE 1.x compatibility)
+                memcpyb(ES, DI, ss, &vbe_info_block, 256);
+	}
                 
-                result = 0x4f;
+        result = 0x4f;
 
-#ifdef VBE2_NO_VESA_CHECK
-#else
-        }
-        else
-        {
-#ifdef DEBUG
-                printf("VBE failed VBE2 signature check\n");
-#endif
-                result = 0x0100;
-        }
-#endif        
         write_word(ss, AX, result);
 }
 
@@ -413,32 +531,30 @@ Bit16u *AX;Bit16u CX; Bit16u ES;Bit16u DI;
         Bit16u            result=0x0100;
         Bit16u            ss=get_SS();
         ModeInfoBlock     info;
-        ModeInfoListItem  *cur_info=&mode_info_list;
-        Boolean           found=0;
+        ModeInfoListItem  *cur_info;
+        Boolean           using_lfb;
 
 #ifdef DEBUG
         printf("VBE vbe_biosfn_return_mode_information ES%x DI%x CX%x\n",ES,DI,CX);
 #endif
 
-        while ((cur_info->mode != VBE_VESA_MODE_END_OF_LIST) && (!found))
-        {
-                if (cur_info->mode == CX)
-                {
-                        found=1;
-                }
-                else
-                {
-                        cur_info++;
-                }
-        }
+        using_lfb=((CX & VBE_MODE_LINEAR_FRAME_BUFFER) == VBE_MODE_LINEAR_FRAME_BUFFER);
         
-        if (found)
+        CX = (CX & 0x1ff);
+        
+        cur_info = mode_info_find_mode(CX, using_lfb, &cur_info);
+
+        if (cur_info != 0)
         {
 #ifdef DEBUG
                 printf("VBE found mode %x\n",CX);
 #endif        
                 memsetb(ss, &info, 0, sizeof(ModeInfoBlock));
                 memcpyb(ss, &info, 0xc000, &(cur_info->info), sizeof(ModeInfoBlockCompact));
+                if (info.WinAAttributes & VBE_WINDOW_ATTRIBUTE_RELOCATABLE) {
+                  info.WinFuncPtr = 0xC0000000UL;
+                  *(Bit16u *)&(info.WinFuncPtr) = (Bit16u)(dispi_set_bank_farcall);
+                }
                 
                 result = 0x4f;
         }
@@ -474,14 +590,17 @@ Bit16u *AX;Bit16u BX; Bit16u ES;Bit16u DI;
 {
         Bit16u            ss = get_SS();
         Bit16u            result;
-        ModeInfoListItem  *cur_info = &mode_info_list;
-        Boolean           found = 0;
+        ModeInfoListItem  *cur_info;
         Boolean           using_lfb;
-        
+        Bit8u             no_clear;
+        Bit8u             lfb_flag;
+
         using_lfb=((BX & VBE_MODE_LINEAR_FRAME_BUFFER) == VBE_MODE_LINEAR_FRAME_BUFFER);
-        
+        lfb_flag=using_lfb?VBE_DISPI_LFB_ENABLED:0;
+        no_clear=((BX & VBE_MODE_PRESERVE_DISPLAY_MEMORY) == VBE_MODE_PRESERVE_DISPLAY_MEMORY)?VBE_DISPI_NOCLEARMEM:0;
+
         BX = (BX & 0x1ff);
-        
+
         //result=read_word(ss,AX);
         
         // check for non vesa mode
@@ -498,18 +617,9 @@ Bit16u *AX;Bit16u BX; Bit16u ES;Bit16u DI;
                 result = 0x4f;
         }
         
-        while ((cur_info->mode != VBE_VESA_MODE_END_OF_LIST) && (!found))
-        {
-                if (cur_info->mode == BX)
-                {
-                        found = 1;
-                }
-                else
-                {
-                        cur_info++;
-                }
-        }
-        if (found)
+        cur_info = mode_info_find_mode(BX, using_lfb, &cur_info);
+
+        if (cur_info != 0)
         {
 #ifdef DEBUG
                 printf("VBE found mode %x, setting:\n", BX);
@@ -518,27 +628,25 @@ Bit16u *AX;Bit16u BX; Bit16u ES;Bit16u DI;
                         cur_info->info.YResolution,
                         cur_info->info.BitsPerPixel);
 #endif
-                // FIXME: this is here so we can do some testing
-                // at least until bochs host side display is up & running
-                // (we're using the 'standard' 320x200x256 vga mode as if it
-                //  were a vesa mode)
                 
-                if (cur_info->info.BitsPerPixel == 8)
-                {
-                  // we have a 8bpp mode, preparing to set it
-                  
-                  dispi_set_xres(cur_info->info.XResolution);
-                  dispi_set_yres(cur_info->info.YResolution);
-                  dispi_set_bpp(VBE_DISPI_BPP_8);
-                  dispi_set_bank(0);
-                  dispi_set_enable(VBE_DISPI_ENABLED);
+                // first disable current mode (when switching between vesa modi)
+                dispi_set_enable(VBE_DISPI_DISABLED);
 
-                  // FIXME: store current mode in CMOS
-                  
-                  result = 0x4f;                  
+                if (cur_info->mode == VBE_VESA_MODE_800X600X4)
+                {
+                  biosfn_set_video_mode(0x6a);
                 }
-                
-                //FIXME: new resolutions will need special code (per bpp)
+
+                dispi_set_xres(cur_info->info.XResolution);
+                dispi_set_yres(cur_info->info.YResolution);
+                dispi_set_bpp(cur_info->info.BitsPerPixel);
+                dispi_set_bank(0);
+                dispi_set_enable(VBE_DISPI_ENABLED | no_clear | lfb_flag);
+
+                write_word(BIOSMEM_SEG,BIOSMEM_VBE_MODE,BX);
+                write_byte(BIOSMEM_SEG,BIOSMEM_VIDEO_CTL,(0x60 | no_clear));
+
+                result = 0x4f;                  
         }
         else
         {
@@ -569,15 +677,22 @@ void vbe_biosfn_return_current_mode(AX, BX)
 Bit16u *AX;Bit16u *BX;
 {
         Bit16u ss=get_SS();
-
-// FIXME: get current mode from CMOS reg?
+        Bit16u mode;
 
 #ifdef DEBUG
         printf("VBE vbe_biosfn_return_current_mode\n");
 #endif
-        
+
+        if(dispi_get_enable())
+        {
+                mode=read_word(BIOSMEM_SEG,BIOSMEM_VBE_MODE);
+        }
+        else
+        {
+                mode=read_byte(BIOSMEM_SEG,BIOSMEM_CURRENT_MODE);
+        }
         write_word(ss, AX, 0x4f);
-        write_word(ss, BX, 0x3);
+        write_word(ss, BX, mode);
 }
 
 
@@ -625,39 +740,148 @@ Bit16u *AX;Bit16u BX;Bit16u *DX;
 {
         Bit16u            ss = get_SS();
         Bit16u            window = read_word(ss, DX);
+        Bit16u            result = 0x014f;
         
         if (BX==0x0000)
         {
                 dispi_set_bank(window);
+                result = 0x4f;
         }
+        else if (BX==0x0100)
+        {
+                window = dispi_get_bank();
+                write_word(ss, DX, result);
+                result = 0x4f;
+        }
+        write_word(ss, AX, result);
 }
 
 
 /** Function 06h - Set/Get Logical Scan Line Length
- * 
+ *
  * Input:
  *              AX      = 4F06h
- * Output:
+ *              BL      = 00h Set Scan Line Length in Pixels
+ *                      = 01h Get Scan Line Length
+ *                      = 02h Set Scan Line Length in Bytes
+ *                      = 03h Get Maximum Scan Line Length
+ *              CX      = If BL=00h Desired Width in Pixels
+ *                        If BL=02h Desired Width in Bytes
+ *                        (Ignored for Get Functions)
+ * 
+ * Output: 
  *              AX      = VBE Return Status
- *
- * FIXME: incomplete API description, Input & Output
+ *              BX      = Bytes Per Scan Line
+ *              CX      = Actual Pixels Per Scan Line
+ *                        (truncated to nearest complete pixel)
+ *              DX      = Maximum Number of Scan Lines 
  */
-void vbe_biosfn_set_get_logical_scan_line_length(AX)
+void vbe_biosfn_set_get_logical_scan_line_length(AX,BX,CX,DX)
+Bit16u *AX;Bit16u *BX;Bit16u *DX;Bit16u *DX;
 {
+	Bit16u ss=get_SS();
+	Bit16u result=0x100;
+	Bit16u width = read_word(ss, CX);
+	Bit16u cmd = read_word(ss, BX);
+	Bit8u  bytespp = dispi_get_bpp()/8;
+
+	// check bl
+	if ( ((cmd & 0xff) == 0x00) || ((cmd & 0xff) == 0x02) )
+	{
+		// set scan line lenght in pixels(0x00) or bytes (0x02)
+		Bit16u new_width;
+		Bit16u new_height;
+
+		if ( ((cmd & 0xff) == 0x02) && (bytespp > 1) )
+		{
+			width/=bytespp;
+		}
+		dispi_set_virt_width(width);
+		new_width=dispi_get_virt_width();
+		new_height=dispi_get_virt_height();
+		
+		if (new_width!=width)
+		{
+#ifdef DEBUG
+                printf("* VBE width adjusted\n");
+#endif
+			
+			// notify width adjusted
+			result=0x024f;
+		}
+		else
+		{
+			result=0x4f;
+		}
+		
+		write_word(ss,BX,new_width*bytespp);
+		write_word(ss,CX,new_width);
+		write_word(ss,DX,new_height);
+	}
+
+	write_word(ss, AX, result);	
 }
 
 
 /** Function 07h - Set/Get Display Start
  * 
- * Input:
+ * Input(16-bit):
  *              AX      = 4F07h
+ *              BH      = 00h Reserved and must be 00h
+ *              BL      = 00h Set Display Start
+ *                      = 01h Get Display Start
+ *                      = 02h Schedule Display Start (Alternate)
+ *                      = 03h Schedule Stereoscopic Display Start
+ *                      = 04h Get Scheduled Display Start Status
+ *                      = 05h Enable Stereoscopic Mode
+ *                      = 06h Disable Stereoscopic Mode
+ *                      = 80h Set Display Start during Vertical Retrace
+ *                      = 82h Set Display Start during Vertical Retrace (Alternate)
+ *                      = 83h Set Stereoscopic Display Start during Vertical Retrace
+ *              ECX     = If BL=02h/82h Display Start Address in bytes
+ *                        If BL=03h/83h Left Image Start Address in bytes
+ *              EDX     = If BL=03h/83h Right Image Start Address in bytes
+ *              CX      = If BL=00h/80h First Displayed Pixel In Scan Line
+ *              DX      = If BL=00h/80h First Displayed Scan Line
+ *
  * Output:
  *              AX      = VBE Return Status
+ *              BH      = If BL=01h Reserved and will be 0
+ *              CX      = If BL=01h First Displayed Pixel In Scan Line
+ *                        If BL=04h 0 if flip has not occurred, not 0 if it has
+ *              DX      = If BL=01h First Displayed Scan Line
  *
- * FIXME: incomplete API description, Input & Output
+ * Input(32-bit): 
+ *              BH      = 00h Reserved and must be 00h
+ *              BL      = 00h Set Display Start
+ *                      = 80h Set Display Start during Vertical Retrace
+ *              CX      = Bits 0-15 of display start address
+ *              DX      = Bits 16-31 of display start address
+ *              ES      = Selector for memory mapped registers 
  */
-void vbe_biosfn_set_get_display_start(AX)
+void vbe_biosfn_set_get_display_start(AX,BX,CX,DX)
+Bit16u *AX;Bit16u BX;Bit16u CX;Bit16u DX;
 {
+	Bit16u ss=get_SS();
+	Bit16u result=0x100;
+#ifdef DEBUG
+ //       printf("VBE vbe_biosfn_set_get_display_start\n");
+#endif
+	
+	// check for set display start
+	if ((GET_BL()==0x00) || (GET_BL()==0x80))
+	{
+	  // 0x80 is during vertical retrace - is also used by sdd vbetest during multibuffering
+#ifdef DEBUG
+//        printf("VBE vbe_biosfn_set_get_display_start CX%x DX%x\n",CX,DX);
+#endif
+		
+		dispi_set_x_offset(CX);
+		dispi_set_y_offset(DX);
+		result = 0x4f;
+	}
+	
+	write_word(ss, AX, result);	
 }
 
 
