@@ -4,7 +4,7 @@
  */
 // ============================================================================================
 //  
-//  Copyright (C) 2001,2002 Christophe Bothamy
+//  Copyright (C) 2001,2002 the LGPL VGABios developers Team
 //
 //  This library is free software; you can redistribute it and/or
 //  modify it under the terms of the GNU Lesser General Public
@@ -140,11 +140,31 @@ static void biosfn_read_video_state_size();
 static void biosfn_save_video_state();
 static void biosfn_restore_video_state();
 
-#asm
+// This is for compiling with gcc2 and gcc3
+#define ASM_START #asm
+#define ASM_END   #endasm
+
+ASM_START
+MACRO SET_INT_VECTOR
+  push ds
+  xor ax, ax
+  mov ds, ax
+  mov ax, ?3
+  mov ?1*4, ax
+  mov ax, ?2
+  mov ?1*4+2, ax
+  pop ds
+MEND
+
+ASM_END
+
+ASM_START
 .text
 .rom
 .org 0
 
+use16 386
+
 vgabios_start:
 .byte	0x55, 0xaa	/* BIOS signature, required for BIOS extensions */
 
@@ -153,10 +173,11 @@ vgabios_start:
 
 vgabios_entry_point:
            
-  JMPL(vgabios_init_func)
+  jmp vgabios_init_func
 
 vgabios_name:
 .ascii	"Plex86/Bochs VGABios"
+.ascii	" "
 .byte	0x00
 
 // Info from Bart Oldeman
@@ -165,29 +186,39 @@ vgabios_name:
 .byte   0x00
 
 vgabios_version:
-.ascii	"v0.3a"
-.byte	0x00
+#ifndef VGABIOS_VERS
+.ascii	"current-cvs"
+#else
+.ascii VGABIOS_VERS
+#endif
+.ascii	" "
 
 vgabios_date:
 .ascii  VGABIOS_DATE
+.byte   0x0a,0x0d
 .byte	0x00
 
 vgabios_copyright:
-.ascii	"(C) 2002 Christophe Bothamy <cbothamy@free.fr>"
+.ascii	"(C) 2003 the LGPL VGABios developers Team"
+.byte	0x0a,0x0d
 .byte	0x00
 
 vgabios_license:
-.ascii	"This VGA Bios is released under the GNU LGPL"
+.ascii	"This VGA/VBE Bios is released under the GNU LGPL"
+.byte	0x0a,0x0d
+.byte	0x0a,0x0d
 .byte	0x00
 
 vgabios_website:
 .ascii	"Please visit :"
 .byte	0x0a,0x0d
-.ascii  " . http://www.plex86.org"
-.byte	0x0a,0x0d
+;;.ascii  " . http://www.plex86.org"
+;;.byte	0x0a,0x0d
 .ascii	" . http://bochs.sourceforge.net"
 .byte	0x0a,0x0d
-.ascii	" . http://savannah.gnu.org/projects/vgabios"
+.ascii	" . http://www.nongnu.org/vgabios"
+.byte	0x0a,0x0d
+.byte	0x0a,0x0d
 .byte	0x00
  
 
@@ -225,6 +256,14 @@ init_no_vbe_display:
   mov ax,#0x0003
   int #0x10
 
+;; copy and activate 8x16 font
+  mov ax, #0x1104
+  mov bl, #0x00
+  int #0x10
+  mov ax, #0x1103
+  mov bl, #0x00
+  int #0x10
+
 ;; show info
   call _display_info
 
@@ -235,12 +274,12 @@ init_no_vbe_display:
 
 
   retf
-#endasm
+ASM_END
 
 /*
  *  int10 handled here
  */
-#asm
+ASM_START
 vgabios_int10_handler:
   pushf
   push es
@@ -257,7 +296,7 @@ vgabios_int10_handler:
   pop es
   popf
   iret
-#endasm
+ASM_END
 
 #include "vgatables.h"
 #include "vgafonts.h"
@@ -267,7 +306,7 @@ vgabios_int10_handler:
  */
 static void init_vga_card()
 {
-#asm
+ASM_START
 ;; switch to color mode and enable CPU access 480 lines
   mov dx, #0x3C2
   mov al, #0xC3
@@ -283,11 +322,9 @@ static void init_vga_card()
   mov al, #0x02
   outb dx,al
 
-#endasm
+ASM_END
 
-#ifdef DEBUG
-  printf("VGABios $Id: vgabios.c,v 1.9 2002/03/10 16:36:06 japj Exp $\n");
-#endif
+  printf("VGABios $Id: vgabios.c,v 1.37 2003/11/04 19:26:43 vruppert Exp $\n");
 }
 
 // --------------------------------------------------------------------------------------------
@@ -330,55 +367,32 @@ static void display_splash_screen()
  *  Tell who we are
  */
 
-#asm
-crlf:
- .byte 0x0a,0x0d,0x00
-space:
- .ascii " "
- .byte 0x00
-#endasm
-
 static void display_info()
 {
-#asm
+ASM_START
  mov ax,#0xc000
  mov ds,ax
  mov si,#vgabios_name
  call _display_string
- mov si,#space
- call _display_string
  mov si,#vgabios_version
  call _display_string
- mov si,#space
- call _display_string
- mov si,#vgabios_date
- call _display_string
- mov si,#crlf
- call _display_string
- mov si,#vgabios_copyright
- call _display_string
- mov si,#crlf
- call _display_string
+ 
+ ;;mov si,#vgabios_copyright
+ ;;call _display_string
+ ;;mov si,#crlf
+ ;;call _display_string
 
  mov si,#vgabios_license
  call _display_string
- mov si,#crlf
- call _display_string
- mov si,#crlf
- call _display_string
  mov si,#vgabios_website
  call _display_string
- mov si,#crlf
- call _display_string
- mov si,#crlf
- call _display_string
-#endasm
+ASM_END
 }
 
 static void display_string()
 {
  // Get length of string
-#asm
+ASM_START
  mov ax,ds
  mov es,ax
  mov di,si
@@ -401,7 +415,7 @@ static void display_string()
  mov bx,#0x000b
  mov bp,si
  int #0x10
-#endasm
+ASM_END
 }
 
 // --------------------------------------------------------------------------------------------
@@ -523,7 +537,7 @@ static void int10_func(DI, SI, BP, SP, BX, DX, CX, AX, DS, ES, FLAGS)
         biosfn_get_all_palette_reg(ES,DX);
         break;
        case 0x10:
-        biosfn_set_single_dac_reg(BX,GET_CH(),GET_CL(),GET_DL());
+        biosfn_set_single_dac_reg(BX,GET_CH(),GET_CL(),GET_DH());
         break;
        case 0x12:
         biosfn_set_all_dac_reg(BX,CX,ES,DX);
@@ -710,49 +724,53 @@ static void int10_func(DI, SI, BP, SP, BX, DX, CX, AX, DS, ES, FLAGS)
 #ifdef DEBUG
           unimplemented();
 #endif
+          // function failed
+          AX=0x100;
           break;
          case 0x05:
           vbe_biosfn_display_window_control(&AX,BX,&DX);
           break;
          case 0x06:
-          //FIXME
-#ifdef DEBUG
-          unimplemented();
-#endif
+          vbe_biosfn_set_get_logical_scan_line_length(&AX,&BX,&CX,&DX);
           break;
          case 0x07:
-          //FIXME
-#ifdef DEBUG   
-          unimplemented();
-#endif
+          vbe_biosfn_set_get_display_start(&AX,BX,CX,DX);
           break;
          case 0x08:
           //FIXME
 #ifdef DEBUG
           unimplemented();
 #endif
+          // function failed
+          AX=0x100;
           break;
          case 0x09:
           //FIXME
 #ifdef DEBUG
           unimplemented();
 #endif
+          // function failed
+          AX=0x100;
           break;
          case 0x0A:
           //FIXME
 #ifdef DEBUG
           unimplemented();
 #endif
+          // function failed
+          AX=0x100;
           break;
 #ifdef DEBUG
          default:
           unknown();
 #endif   		 
+          // function failed
+          AX=0x100;
           }
         }
         else {
           // No VBE display
-          SET_AH(0x01);
+          AX=0x0100;
           }
         break;
 #endif
@@ -775,7 +793,7 @@ static void biosfn_set_video_mode(mode) Bit8u mode;
 
  // Should we clear the screen ?
  Bit8u noclearmem=mode&0x80;
- Bit8u line,*palette;
+ Bit8u line,mmask,*palette;
  Bit16u i,twidth,theight,cheight;
  Bit8u modeset_ctl,video_ctl,vga_switches;
  Bit16u crtc_addr;
@@ -854,6 +872,9 @@ static void biosfn_set_video_mode(mode) Bit8u mode;
     }
   }
 
+ // Reset Attribute Ctl flip-flop
+ inb(VGAREG_ACTL_RESET);
+
  // Set Attribute Ctl
  for(i=0;i<=ACTL_MAX_REG;i++)
   {outb(VGAREG_ACTL_ADDRESS,i);
@@ -889,17 +910,30 @@ static void biosfn_set_video_mode(mode) Bit8u mode;
  inb(VGAREG_ACTL_RESET);
 
  if(noclearmem==0x00)
-  {if(vga_modes[line].class==TEXT)
+  {
+   if(vga_modes[line].class==TEXT)
     {
      memsetw(vga_modes[line].sstart,0,0x0720,0x4000); // 32k
     }
    else
-    {// FIXME should handle gfx mode
+    {
+     if(mode<0x0d)
+      {
+       memsetw(vga_modes[line].sstart,0,0x0000,0x4000); // 32k
+      }
+     else
+      {
+       outb( VGAREG_SEQU_ADDRESS, 0x02 );
+       mmask = inb( VGAREG_SEQU_DATA );
+       outb( VGAREG_SEQU_DATA, 0x0f ); // all planes
+       memsetw(vga_modes[line].sstart,0,0x0000,0x8000); // 64k
+       outb( VGAREG_SEQU_DATA, mmask );
+      }
     }
   }
 
  // Set the BIOS mem
- write_byte(BIOSMEM_SEG,BIOSMEM_CURRENT_MODE,mode|noclearmem);
+ write_byte(BIOSMEM_SEG,BIOSMEM_CURRENT_MODE,mode);
  write_word(BIOSMEM_SEG,BIOSMEM_NB_COLS,twidth);
  write_word(BIOSMEM_SEG,BIOSMEM_PAGE_SIZE,vga_modes[line].slength);
  write_word(BIOSMEM_SEG,BIOSMEM_CRTC_ADDRESS,crtc_addr);
@@ -921,8 +955,7 @@ static void biosfn_set_video_mode(mode) Bit8u mode;
  // Set cursor shape
  if(vga_modes[line].class==TEXT)
   {
-    //biosfn_set_cursor_shape(0x06,0x07); // CGA
-    biosfn_set_cursor_shape(0x0A,0x0B); // Monochrome
+   biosfn_set_cursor_shape(0x06,0x07);
   }
 
  // Set cursor pos for page 0..7
@@ -933,28 +966,31 @@ static void biosfn_set_video_mode(mode) Bit8u mode;
  biosfn_set_active_page(0x00);
 
  // Write the fonts in memory
-// FIXME
+ if(vga_modes[line].class==TEXT)
+  {
+   biosfn_load_text_8_16_pat(0x04,0x00);
+  }
 
  // Set the ints 0x1F and 0x43
-#asm
- SET_INT_VECTOR(0x1f, #0xC000, #vgafont8+128*8)
-#endasm
+ASM_START
+ SET_INT_VECTOR(0x1f, #0xC000, #_vgafont8+128*8)
+ASM_END
 
   switch(cheight)
    {case 8:
-#asm
-     SET_INT_VECTOR(0x43, #0xC000, #vgafont8)
-#endasm
+ASM_START
+     SET_INT_VECTOR(0x43, #0xC000, #_vgafont8)
+ASM_END
      break;
     case 14:
-#asm
-     SET_INT_VECTOR(0x43, #0xC000, #vgafont14)
-#endasm
+ASM_START
+     SET_INT_VECTOR(0x43, #0xC000, #_vgafont14)
+ASM_END
      break;
     case 16:
-#asm
-     SET_INT_VECTOR(0x43, #0xC000, #vgafont16)
-#endasm
+ASM_START
+     SET_INT_VECTOR(0x43, #0xC000, #_vgafont16)
+ASM_END
      break;
    }
 }
@@ -962,20 +998,34 @@ static void biosfn_set_video_mode(mode) Bit8u mode;
 // --------------------------------------------------------------------------------------------
 static void biosfn_set_cursor_shape (CH,CL) 
 Bit8u CH;Bit8u CL; 
-{Bit16u curs;
+{Bit16u cheight,curs,crtc_addr;
 
  CH&=0x3f;
  CL&=0x1f;
 
- // FIXME should remap to vga settings in case of 14/16 lines
  curs=(CH<<8)+CL;
  write_word(BIOSMEM_SEG,BIOSMEM_CURSOR_TYPE,curs);
 
+ cheight = read_word(BIOSMEM_SEG,BIOSMEM_CHAR_HEIGHT);
+ if((cheight>8) && (CL<8) && (CH<0x20))
+  {
+   if(CL!=(CH+1))
+    {
+     CH = ((CH+1) * cheight / 8) -1;
+    }
+   else
+    {
+     CH = ((CL+1) * cheight / 8) - 2;
+    }
+   CL = ((CL+1) * cheight / 8) - 1;
+  }
+
  // CTRC regs 0x0a and 0x0b
- outb(read_word(BIOSMEM_SEG,BIOSMEM_CRTC_ADDRESS),0x0a);
- outb(read_word(BIOSMEM_SEG,BIOSMEM_CRTC_ADDRESS)+1,CH);
- outb(read_word(BIOSMEM_SEG,BIOSMEM_CRTC_ADDRESS),0x0b);
- outb(read_word(BIOSMEM_SEG,BIOSMEM_CRTC_ADDRESS)+1,CL);
+ crtc_addr=read_word(BIOSMEM_SEG,BIOSMEM_CRTC_ADDRESS);
+ outb(crtc_addr,0x0a);
+ outb(crtc_addr+1,CH);
+ outb(crtc_addr,0x0b);
+ outb(crtc_addr+1,CL);
 }
 
 // --------------------------------------------------------------------------------------------
@@ -983,7 +1033,7 @@ static void biosfn_set_cursor_pos (page, cursor)
 Bit8u page;Bit16u cursor;
 {
  Bit8u xcurs,ycurs,current;
- Bit16u nbcols,nbrows,address;
+ Bit16u nbcols,nbrows,address,crtc_addr;
 
  // Should not happen...
  if(page>7)return;
@@ -1005,10 +1055,11 @@ Bit8u page;Bit16u cursor;
    address=SCREEN_IO_START(nbcols,nbrows,page)+xcurs+ycurs*nbcols;
    
    // CRTC regs 0x0e and 0x0f
-   outb(read_word(BIOSMEM_SEG,BIOSMEM_CRTC_ADDRESS),0x0e);
-   outb(read_word(BIOSMEM_SEG,BIOSMEM_CRTC_ADDRESS)+1,(address&0xff00)>>8);
-   outb(read_word(BIOSMEM_SEG,BIOSMEM_CRTC_ADDRESS),0x0f);
-   outb(read_word(BIOSMEM_SEG,BIOSMEM_CRTC_ADDRESS)+1,address&0x00ff);
+   crtc_addr=read_word(BIOSMEM_SEG,BIOSMEM_CRTC_ADDRESS);
+   outb(crtc_addr,0x0e);
+   outb(crtc_addr+1,(address&0xff00)>>8);
+   outb(crtc_addr,0x0f);
+   outb(crtc_addr+1,address&0x00ff);
   }
 }
 
@@ -1032,30 +1083,44 @@ Bit8u page;Bit16u *shape;Bit16u *pos;
 static void biosfn_set_active_page (page) 
 Bit8u page;
 {
- Bit16u cursor,dummy;
+ Bit16u cursor,dummy,crtc_addr;
  Bit16u nbcols,nbrows,address;
+ Bit8u mode,line;
 
  if(page>7)return;
 
+ // Get the mode
+ mode=read_byte(BIOSMEM_SEG,BIOSMEM_CURRENT_MODE);
+ line=find_vga_entry(mode);
+ if(line==0xFF)return;
+
  // Get pos curs pos for the right page 
  biosfn_get_cursor_pos(page,&dummy,&cursor);
 
- // Get the dimensions
- nbcols=read_word(BIOSMEM_SEG,BIOSMEM_NB_COLS);
- nbrows=read_byte(BIOSMEM_SEG,BIOSMEM_NB_ROWS)+1;
+ if(vga_modes[line].class==TEXT)
+  {
+   // Get the dimensions
+   nbcols=read_word(BIOSMEM_SEG,BIOSMEM_NB_COLS);
+   nbrows=read_byte(BIOSMEM_SEG,BIOSMEM_NB_ROWS)+1;
  
- // Calculate the address knowing nbcols nbrows and page num
- address=SCREEN_MEM_START(nbcols,nbrows,page);
- write_word(BIOSMEM_SEG,BIOSMEM_CURRENT_START,address);
+   // Calculate the address knowing nbcols nbrows and page num
+   address=SCREEN_MEM_START(nbcols,nbrows,page);
+   write_word(BIOSMEM_SEG,BIOSMEM_CURRENT_START,address);
 
- // Start address
- address=SCREEN_IO_START(nbcols,nbrows,page);
+   // Start address
+   address=SCREEN_IO_START(nbcols,nbrows,page);
+  }
+ else
+  {
+   address = page*vga_modes[line].slength;
+  }
 
  // CRTC regs 0x0c and 0x0d
- outb(read_word(BIOSMEM_SEG,BIOSMEM_CRTC_ADDRESS),0x0c);
- outb(read_word(BIOSMEM_SEG,BIOSMEM_CRTC_ADDRESS)+1,(address&0xff00)>>8);
- outb(read_word(BIOSMEM_SEG,BIOSMEM_CRTC_ADDRESS),0x0d);
- outb(read_word(BIOSMEM_SEG,BIOSMEM_CRTC_ADDRESS)+1,address&0x00ff);
+ crtc_addr=read_word(BIOSMEM_SEG,BIOSMEM_CRTC_ADDRESS);
+ outb(crtc_addr,0x0c);
+ outb(crtc_addr+1,(address&0xff00)>>8);
+ outb(crtc_addr,0x0d);
+ outb(crtc_addr+1,address&0x00ff);
 
  // And change the BIOS page
  write_byte(BIOSMEM_SEG,BIOSMEM_CURRENT_PAGE,page);
@@ -1094,41 +1159,52 @@ Bit8u nblines;Bit8u attr;Bit8u rul;Bit8u cul;Bit8u rlr;Bit8u clr;Bit8u page;Bit8
  if(page==0xFF)
   page=read_byte(BIOSMEM_SEG,BIOSMEM_CURRENT_PAGE);
 
- // Compute the address
- address=SCREEN_MEM_START(nbcols,nbrows,page);
+ if(vga_modes[line].class==TEXT)
+  {
+   // Compute the address
+   address=SCREEN_MEM_START(nbcols,nbrows,page);
 #ifdef DEBUG
- printf("Scroll, address %04x (%04x %04x %02x)\n",address,nbrows,nbcols,page);
+   printf("Scroll, address %04x (%04x %04x %02x)\n",address,nbrows,nbcols,page);
 #endif
 
- if(rlr>=nbrows)rlr=nbrows-1;
- if(clr>=nbcols)clr=nbcols-1;
- if(nblines>nbrows)nblines=0;
+   if(rlr>=nbrows)rlr=nbrows-1;
+   if(clr>=nbcols)clr=nbcols-1;
+   if(nblines>nbrows)nblines=0;
 
- if(nblines==0&&rul==0&&cul==0&&rlr==nbrows-1&&clr==nbcols-1)
-  {
-   memsetw(vga_modes[line].sstart,address,(Bit16u)attr*0x100+' ',nbrows*nbcols);
-  }
- else
-  {// if Scroll up
-   if(dir==SCROLL_UP)
-    {for(i=rul;i<=rlr;i++)
-      {
-       if((i+nblines>rlr)||(nblines==0))
-        memsetw(vga_modes[line].sstart,address+(i*nbcols+cul)*2,(Bit16u)attr*0x100+' ',clr-cul+1);
-       else
-        memcpyw(vga_modes[line].sstart,address+(i*nbcols+cul)*2,vga_modes[line].sstart,((i+nblines)*nbcols+cul)*2,clr-cul+1);
-      }
+   if(nblines==0&&rul==0&&cul==0&&rlr==nbrows-1&&clr==nbcols-1)
+    {
+     memsetw(vga_modes[line].sstart,address,(Bit16u)attr*0x100+' ',nbrows*nbcols);
     }
    else
-    {for(i=rlr;i>=rul;i--)
-      {
-       if((i<rul+nblines)||(nblines==0))
-        memsetw(vga_modes[line].sstart,address+(i*nbcols+cul)*2,(Bit16u)attr*0x100+' ',clr-cul+1);
-       else
-        memcpyw(vga_modes[line].sstart,address+(i*nbcols+cul)*2,vga_modes[line].sstart,((i-nblines)*nbcols+cul)*2,clr-cul+1);
+    {// if Scroll up
+     if(dir==SCROLL_UP)
+      {for(i=rul;i<=rlr;i++)
+        {
+         if((i+nblines>rlr)||(nblines==0))
+          memsetw(vga_modes[line].sstart,address+(i*nbcols+cul)*2,(Bit16u)attr*0x100+' ',clr-cul+1);
+         else
+          memcpyw(vga_modes[line].sstart,address+(i*nbcols+cul)*2,vga_modes[line].sstart,((i+nblines)*nbcols+cul)*2,clr-cul+1);
+        }
+      }
+     else
+      {for(i=rlr;i>=rul;i--)
+        {
+         if((i<rul+nblines)||(nblines==0))
+          memsetw(vga_modes[line].sstart,address+(i*nbcols+cul)*2,(Bit16u)attr*0x100+' ',clr-cul+1);
+         else
+          memcpyw(vga_modes[line].sstart,address+(i*nbcols+cul)*2,vga_modes[line].sstart,((i-nblines)*nbcols+cul)*2,clr-cul+1);
+        }
       }
     }
   }
+ else
+  {
+   // FIXME gfx mode
+#ifdef DEBUG
+   printf("Scroll in graphics mode !\n");
+   unimplemented();
+#endif
+  }
 }
 
 // --------------------------------------------------------------------------------------------
@@ -1139,8 +1215,6 @@ Bit8u page;Bit16u *car;
  Bit16u nbcols,nbrows,address;
  Bit16u cursor,dummy;
 
- // FIXME gfx mode
-
  // Get the mode
  mode=read_byte(BIOSMEM_SEG,BIOSMEM_CURRENT_MODE);
  line=find_vga_entry(mode);
@@ -1154,22 +1228,63 @@ Bit8u page;Bit16u *car;
  nbrows=read_byte(BIOSMEM_SEG,BIOSMEM_NB_ROWS)+1;
  nbcols=read_word(BIOSMEM_SEG,BIOSMEM_NB_COLS);
 
- // Compute the address
- address=SCREEN_MEM_START(nbcols,nbrows,page)+(xcurs+ycurs*nbcols)*2;
+ if(vga_modes[line].class==TEXT)
+  {
+   // Compute the address
+   address=SCREEN_MEM_START(nbcols,nbrows,page)+(xcurs+ycurs*nbcols)*2;
 
- write_word(ss,car,read_word(vga_modes[line].sstart,address));
+   write_word(ss,car,read_word(vga_modes[line].sstart,address));
+  }
+ else
+  {
+   // FIXME gfx mode
+#ifdef DEBUG
+   unimplemented();
+#endif
+  }
+}
+
+// --------------------------------------------------------------------------------------------
+static void write_gfx_char(car,attr,xcurs,ycurs,nbcols,cheight)
+Bit8u car;Bit8u attr;Bit8u xcurs;Bit8u ycurs;Bit8u nbcols;Bit8u cheight;
+{
+ Bit8u i,mmask;
+ Bit8u *fdata;
+ Bit16u addr,dest,src;
+
+ switch(cheight)
+  {case 14:
+    fdata = &vgafont14;
+    break;
+   case 16:
+    fdata = &vgafont16;
+    break;
+   default:
+    fdata = &vgafont8;
+  }
+ addr=xcurs+ycurs*cheight*nbcols;
+ src = car * cheight;
+ outb( VGAREG_SEQU_ADDRESS, 0x02 );
+ mmask = inb( VGAREG_SEQU_DATA );
+ for(i=0;i<cheight;i++)
+  {
+   dest=addr+i*nbcols;
+   outb( VGAREG_SEQU_DATA, 0x0f );
+   write_byte(0xa000,dest,0x00);
+   outb( VGAREG_SEQU_DATA, attr & 0x0f );
+   write_byte(0xa000,dest,fdata[src+i]);
+  }
+ outb( VGAREG_SEQU_DATA, mmask );
 }
 
 // --------------------------------------------------------------------------------------------
 static void biosfn_write_char_attr (car,page,attr,count) 
 Bit8u car;Bit8u page;Bit8u attr;Bit16u count;
 {
- Bit8u xcurs,ycurs,mode,line;
+ Bit8u cheight,xcurs,ycurs,mode,line;
  Bit16u nbcols,nbrows,address;
  Bit16u cursor,dummy;
 
- // FIXME gfx mode
-
  // Get the mode
  mode=read_byte(BIOSMEM_SEG,BIOSMEM_CURRENT_MODE);
  line=find_vga_entry(mode);
@@ -1183,23 +1298,43 @@ Bit8u car;Bit8u page;Bit8u attr;Bit16u count;
  nbrows=read_byte(BIOSMEM_SEG,BIOSMEM_NB_ROWS)+1;
  nbcols=read_word(BIOSMEM_SEG,BIOSMEM_NB_COLS);
 
- // Compute the address
- address=SCREEN_MEM_START(nbcols,nbrows,page)+(xcurs+ycurs*nbcols)*2;
+ if(vga_modes[line].class==TEXT)
+  {
+   // Compute the address
+   address=SCREEN_MEM_START(nbcols,nbrows,page)+(xcurs+ycurs*nbcols)*2;
 
- dummy=((Bit16u)attr<<8)+car;
- memsetw(vga_modes[line].sstart,address,dummy,count);
+   dummy=((Bit16u)attr<<8)+car;
+   memsetw(vga_modes[line].sstart,address,dummy,count);
+  }
+ else
+  {
+   // FIXME gfx mode not complete
+   if(vga_modes[line].memmodel==PLANAR4)
+    {
+     cheight=vga_modes[line].cheight;
+     while((count-->0) && (xcurs<nbcols))
+      {
+       write_gfx_char(car,attr,xcurs,ycurs,nbcols,cheight);
+       xcurs++;
+      }
+    }
+   else
+    {
+#ifdef DEBUG
+     unimplemented();
+#endif
+    }
+  }
 }
 
 // --------------------------------------------------------------------------------------------
 static void biosfn_write_char_only (car,page,attr,count)
 Bit8u car;Bit8u page;Bit8u attr;Bit16u count;
 {
- Bit8u xcurs,ycurs,mode,line;
+ Bit8u cheight,xcurs,ycurs,mode,line;
  Bit16u nbcols,nbrows,address;
  Bit16u cursor,dummy;
 
- // FIXME gfx mode
-
  // Get the mode
  mode=read_byte(BIOSMEM_SEG,BIOSMEM_CURRENT_MODE);
  line=find_vga_entry(mode);
@@ -1213,12 +1348,34 @@ Bit8u car;Bit8u page;Bit8u attr;Bit16u count;
  nbrows=read_byte(BIOSMEM_SEG,BIOSMEM_NB_ROWS)+1;
  nbcols=read_word(BIOSMEM_SEG,BIOSMEM_NB_COLS);
 
- // Compute the address
- address=SCREEN_MEM_START(nbcols,nbrows,page)+(xcurs+ycurs*nbcols)*2;
+ if(vga_modes[line].class==TEXT)
+  {
+   // Compute the address
+   address=SCREEN_MEM_START(nbcols,nbrows,page)+(xcurs+ycurs*nbcols)*2;
 
- while(count-->0)
-  {write_byte(vga_modes[line].sstart,address,car);
-   address+=2;
+   while(count-->0)
+    {write_byte(vga_modes[line].sstart,address,car);
+     address+=2;
+    }
+  }
+ else
+  {
+   // FIXME gfx mode not complete
+   if(vga_modes[line].memmodel==PLANAR4)
+    {
+     cheight=vga_modes[line].cheight;
+     while((count-->0) && (xcurs<nbcols))
+      {
+       write_gfx_char(car,attr,xcurs,ycurs,nbcols,cheight);
+       xcurs++;
+      }
+    }
+   else
+    {
+#ifdef DEBUG
+     unimplemented();
+#endif
+    }
   }
 }
 
@@ -1263,7 +1420,7 @@ static void biosfn_write_teletype (car, page, attr, flag)
 Bit8u car;Bit8u page;Bit8u attr;Bit8u flag;
 {// flag = WITH_ATTR / NO_ATTR
 
- Bit8u xcurs,ycurs,mode,line;
+ Bit8u cheight,xcurs,ycurs,mode,line;
  Bit16u nbcols,nbrows,address;
  Bit16u cursor,dummy;
 
@@ -1315,14 +1472,33 @@ Bit8u car;Bit8u page;Bit8u attr;Bit8u flag;
     break;
 
    default:
-    // Compute the address  
-    address=SCREEN_MEM_START(nbcols,nbrows,page)+(xcurs+ycurs*nbcols)*2;
 
-    // Write the char 
-    write_byte(vga_modes[line].sstart,address,car);
+    if(vga_modes[line].class==TEXT)
+     {
+      // Compute the address  
+      address=SCREEN_MEM_START(nbcols,nbrows,page)+(xcurs+ycurs*nbcols)*2;
 
-    if(flag==WITH_ATTR)
-     write_byte(vga_modes[line].sstart,address+1,attr);
+      // Write the char 
+      write_byte(vga_modes[line].sstart,address,car);
+
+      if(flag==WITH_ATTR)
+       write_byte(vga_modes[line].sstart,address+1,attr);
+     }
+    else
+     {
+      // FIXME gfx mode not complete
+      if(vga_modes[line].memmodel==PLANAR4)
+       {
+        cheight=vga_modes[line].cheight;
+        write_gfx_char(car,attr,xcurs,ycurs,nbcols,cheight);
+       }
+      else
+       {
+#ifdef DEBUG
+        unimplemented();
+#endif
+       }
+     }
 
     xcurs++;
   }
@@ -1348,14 +1524,15 @@ Bit8u car;Bit8u page;Bit8u attr;Bit8u flag;
 static void biosfn_get_video_mode (AX,BX) 
 Bit16u *AX;Bit16u *BX;
 {Bit16u ss=get_SS();
- Bit8u  mode,page;
+ Bit8u  mode,noclear,page;
  Bit16u nbcars;
 
  page=read_byte(BIOSMEM_SEG,BIOSMEM_CURRENT_PAGE);
  mode=read_byte(BIOSMEM_SEG,BIOSMEM_CURRENT_MODE);
+ noclear=read_byte(BIOSMEM_SEG,BIOSMEM_VIDEO_CTL) & 0x80;
  nbcars=read_word(BIOSMEM_SEG,BIOSMEM_NB_COLS);
 
- write_word(ss,AX,(nbcars<<8)+mode);
+ write_word(ss,AX,(nbcars<<8)+mode|noclear);
  write_word(ss,BX,((Bit16u)page)<<8);
 }
 
@@ -1368,6 +1545,7 @@ Bit8u reg;Bit8u value;
    inb(VGAREG_ACTL_RESET);
    outb(VGAREG_ACTL_ADDRESS,reg);
    outb(VGAREG_ACTL_WRITE_DATA,value);
+   outb(VGAREG_ACTL_ADDRESS,0x20);
   }
 }
 
@@ -1378,6 +1556,7 @@ Bit8u value;
  inb(VGAREG_ACTL_RESET);
  outb(VGAREG_ACTL_ADDRESS,0x11);
  outb(VGAREG_ACTL_WRITE_DATA,value);
+ outb(VGAREG_ACTL_ADDRESS,0x20);
 }
 
 // --------------------------------------------------------------------------------------------
@@ -1388,7 +1567,7 @@ Bit16u seg;Bit16u offset;
 
  inb(VGAREG_ACTL_RESET);
  // First the colors
- for(i=0;i<=0x10;i++)
+ for(i=0;i<0x10;i++)
   {
    outb(VGAREG_ACTL_ADDRESS,i);
    outb(VGAREG_ACTL_WRITE_DATA,read_byte(seg,offset));
@@ -1398,6 +1577,7 @@ Bit16u seg;Bit16u offset;
  // Then the border
  outb(VGAREG_ACTL_ADDRESS,0x11);
  outb(VGAREG_ACTL_WRITE_DATA,read_byte(seg,offset));
+ outb(VGAREG_ACTL_ADDRESS,0x20);
 }
 
 // --------------------------------------------------------------------------------------------
@@ -1411,10 +1591,8 @@ Bit8u state;
  value=inb(VGAREG_ACTL_READ_DATA);
  value&=0xf7;
  value|=state<<3;
-
- inb(VGAREG_ACTL_RESET);
- outb(VGAREG_ACTL_ADDRESS,0x10);
  outb(VGAREG_ACTL_WRITE_DATA,value);
+ outb(VGAREG_ACTL_ADDRESS,0x20);
 }
 
 // --------------------------------------------------------------------------------------------
@@ -1427,6 +1605,8 @@ Bit8u reg;Bit16u *value;
    inb(VGAREG_ACTL_RESET);
    outb(VGAREG_ACTL_ADDRESS,reg);
    write_word(ss,value,((Bit16u)inb(VGAREG_ACTL_READ_DATA))<<8);
+   inb(VGAREG_ACTL_RESET);
+   outb(VGAREG_ACTL_ADDRESS,0x20);
   }
 }
 
@@ -1438,6 +1618,7 @@ Bit16u *value;
  inb(VGAREG_ACTL_RESET);
  outb(VGAREG_ACTL_ADDRESS,0x11);
  write_word(ss,value,((Bit16u)inb(VGAREG_ACTL_READ_DATA))<<8);
+ outb(VGAREG_ACTL_ADDRESS,0x20);
 }
 
 // --------------------------------------------------------------------------------------------
@@ -1445,10 +1626,10 @@ static void biosfn_get_all_palette_reg (seg,offset) Bit16u seg;Bit16u offset;
 {
  Bit8u i;
 
- inb(VGAREG_ACTL_RESET);
  // First the colors
  for(i=0;i<=0x10;i++)
   {
+   inb(VGAREG_ACTL_RESET);
    outb(VGAREG_ACTL_ADDRESS,i);
    write_byte(seg,offset,inb(VGAREG_ACTL_READ_DATA));
    offset++;
@@ -1457,6 +1638,7 @@ static void biosfn_get_all_palette_reg (seg,offset) Bit16u seg;Bit16u offset;
  // Then the border
  outb(VGAREG_ACTL_ADDRESS,0x11);
  write_byte(seg,offset,inb(VGAREG_ACTL_READ_DATA));
+ outb(VGAREG_ACTL_ADDRESS,0x20);
 }
 
 // --------------------------------------------------------------------------------------------
@@ -1472,7 +1654,7 @@ Bit16u reg;Bit8u g;Bit8u b;Bit8u r;
 // --------------------------------------------------------------------------------------------
 static void biosfn_set_all_dac_reg (start,count,seg,offset) 
 Bit16u start;Bit16u count;Bit16u seg;Bit16u offset;
-{Bit8u i;
+{Bit16u i;
  outb(VGAREG_DAC_WRITE_ADDRESS,start);
  for(i=0;i<count;i++)
   {outb(VGAREG_DAC_DATA,read_byte(seg,offset++));
@@ -1492,16 +1674,13 @@ Bit8u function;
  value=inb(VGAREG_ACTL_READ_DATA);
  function&=0x01;
  if(function==0)
-  {// 4 of 64
+  {// set paging code
    value&=0x7f;
-   value|=function<<7;
-
-   inb(VGAREG_ACTL_RESET);
-   outb(VGAREG_ACTL_ADDRESS,0x10);
+   value|=page<<7;
    outb(VGAREG_ACTL_WRITE_DATA,value);
   }
  else
-  {// 16 of 16
+  {// select page
    inb(VGAREG_ACTL_RESET);
    outb(VGAREG_ACTL_ADDRESS,0x14);
    if(value&0x80)
@@ -1509,6 +1688,7 @@ Bit8u function;
    else
      outb(VGAREG_ACTL_WRITE_DATA,(page&0x03)<<2);
   }
+ outb(VGAREG_ACTL_ADDRESS,0x20);
 }
 
 // --------------------------------------------------------------------------------------------
@@ -1528,7 +1708,7 @@ Bit8u reg;Bit16u *tor;Bit16u *togb;
 // --------------------------------------------------------------------------------------------
 static void biosfn_read_all_dac_reg (start,count,seg,offset) 
 Bit16u start;Bit16u count;Bit16u seg;Bit16u offset;
-{Bit8u i;
+{Bit16u i;
  outb(VGAREG_DAC_READ_ADDRESS,start);
  for(i=0;i<count;i++)
   {write_byte(seg,offset++,inb(VGAREG_DAC_DATA));
@@ -1566,13 +1746,15 @@ static void biosfn_read_video_dac_state (state) Bit16u *state;
  if(mcr==0)(csr>>2)&0x03;
 
  write_word(ss,state,(mcr<<8)+csr);
+ outb(VGAREG_ACTL_ADDRESS,0x20);
 }
 
 // --------------------------------------------------------------------------------------------
 static void biosfn_perform_gray_scale_summing (start,count) 
 Bit16u start;Bit16u count;
-{Bit8u index,r,g,b,d;
+{Bit8u r,g,b,d;
  Bit16u i,m;
+ Bit16u index;
 
  inb(VGAREG_ACTL_RESET);
  outb(VGAREG_ACTL_ADDRESS,0x10);
@@ -1604,39 +1786,151 @@ Bit16u start;Bit16u count;
    outb( VGAREG_DAC_DATA, i&0xff );
    outb( VGAREG_DAC_DATA, i&0xff );
   }  
+ outb(VGAREG_ACTL_ADDRESS,0x20);
 }
 
 // --------------------------------------------------------------------------------------------
+static void get_font_access()
+{
+ outw( VGAREG_SEQU_ADDRESS, 0x0100 );
+ outw( VGAREG_SEQU_ADDRESS, 0x0402 );
+ outw( VGAREG_SEQU_ADDRESS, 0x0704 );
+ outw( VGAREG_SEQU_ADDRESS, 0x0300 );
+ outw( VGAREG_GRDC_ADDRESS, 0x0204 );
+ outw( VGAREG_GRDC_ADDRESS, 0x0005 );
+ outw( VGAREG_GRDC_ADDRESS, 0x0406 );
+}
+
+static void release_font_access()
+{
+ outw( VGAREG_SEQU_ADDRESS, 0x0100 );
+ outw( VGAREG_SEQU_ADDRESS, 0x0302 );
+ outw( VGAREG_SEQU_ADDRESS, 0x0304 );
+ outw( VGAREG_SEQU_ADDRESS, 0x0300 );
+ outw( VGAREG_GRDC_ADDRESS, 0x0004 );
+ outw( VGAREG_GRDC_ADDRESS, 0x1005 );
+ outw( VGAREG_GRDC_ADDRESS, 0x0e06 );
+}
+
+ASM_START
+idiv_u:
+  xor dx,dx
+  div bx
+  ret
+ASM_END
+
+static void set_scan_lines(lines) Bit8u lines;
+{
+ Bit16u crtc_addr,cols,page,vde;
+ Bit8u crtc_r9,ovl,rows;
+
+ crtc_addr = read_word(BIOSMEM_SEG,BIOSMEM_CRTC_ADDRESS);
+ outb(crtc_addr, 0x09);
+ crtc_r9 = inb(crtc_addr+1);
+ crtc_r9 = (crtc_r9 & 0xe0) | (lines - 1);
+ outb(crtc_addr+1, crtc_r9);
+ if(lines==8)
+  {
+   biosfn_set_cursor_shape(0x06,0x07);
+  }
+ else
+  {
+   biosfn_set_cursor_shape(lines-4,lines-3);
+  }
+ write_word(BIOSMEM_SEG,BIOSMEM_CHAR_HEIGHT, lines);
+ outb(crtc_addr, 0x12);
+ vde = inb(crtc_addr+1);
+ outb(crtc_addr, 0x07);
+ ovl = inb(crtc_addr+1);
+ vde += (((ovl & 0x02) << 7) + ((ovl & 0x40) << 3) + 1);
+ rows = vde / lines;
+ write_byte(BIOSMEM_SEG,BIOSMEM_NB_ROWS, rows-1);
+ cols = read_word(BIOSMEM_SEG,BIOSMEM_NB_COLS);
+ write_word(BIOSMEM_SEG,BIOSMEM_PAGE_SIZE, rows * cols * 2);
+}
+
 static void biosfn_load_text_user_pat (AL,ES,BP,CX,DX,BL,BH) Bit8u AL;Bit16u ES;Bit16u BP;Bit16u CX;Bit16u DX;Bit8u BL;Bit8u BH;
 {
-#ifdef DEBUG
- unimplemented();
-#endif
+ Bit16u blockaddr,dest,i,src;
+
+ get_font_access();
+ blockaddr = BL << 13;
+ for(i=0;i<CX;i++)
+  {
+   src = BP + i * BH;
+   dest = blockaddr + (DX + i) * 32;
+   memcpyb(0xA000, dest, ES, src, BH);
+  }
+ release_font_access();
+ if(AL>=0x10)
+  {
+   set_scan_lines(BH);
+  }
 }
+
 static void biosfn_load_text_8_14_pat (AL,BL) Bit8u AL;Bit8u BL;
 {
-#ifdef DEBUG
- unimplemented();
-#endif
+ Bit16u blockaddr,dest,i,src;
+
+ get_font_access();
+ blockaddr = BL << 13;
+ for(i=0;i<0x100;i++)
+  {
+   src = i * 14;
+   dest = blockaddr + i * 32;
+   memcpyb(0xA000, dest, 0xC000, vgafont14+src, 14);
+  }
+ release_font_access();
+ if(AL>=0x10)
+  {
+   set_scan_lines(14);
+  }
 }
+
 static void biosfn_load_text_8_8_pat (AL,BL) Bit8u AL;Bit8u BL;
 {
-#ifdef DEBUG
- unimplemented();
-#endif
+ Bit16u blockaddr,dest,i,src;
+
+ get_font_access();
+ blockaddr = BL << 13;
+ for(i=0;i<0x100;i++)
+  {
+   src = i * 8;
+   dest = blockaddr + i * 32;
+   memcpyb(0xA000, dest, 0xC000, vgafont8+src, 8);
+  }
+ release_font_access();
+ if(AL>=0x10)
+  {
+   set_scan_lines(8);
+  }
 }
+
 static void biosfn_set_text_block_specifier (BL) Bit8u BL;
 {
-#ifdef DEBUG
- unimplemented();
-#endif
+ outb( VGAREG_SEQU_ADDRESS, 0x03 );
+ outb( VGAREG_SEQU_DATA, BL );
 }
+
 static void biosfn_load_text_8_16_pat (AL,BL) Bit8u AL;Bit8u BL;
 {
-#ifdef DEBUG
- unimplemented();
-#endif
+ Bit16u blockaddr,dest,i,src;
+
+ get_font_access();
+ blockaddr = BL << 13;
+ for(i=0;i<0x100;i++)
+  {
+   src = i * 16;
+   dest = blockaddr + i * 32;
+   memcpyb(0xA000, dest, 0xC000, vgafont16+src, 16);
+  }
+ release_font_access();
+ if(AL>=0x10)
+  {
+   set_scan_lines(16);
+  }
 }
+
 static void biosfn_load_gfx_8_8_chars (ES,BP) Bit16u ES;Bit16u BP;
 {
 #ifdef DEBUG
@@ -1712,7 +2006,7 @@ Bit8u BH;Bit16u *ES;Bit16u *BP;Bit16u *CX;Bit16u *DX;
     return;
   }
  // Set byte/char of on screen font
- write_word(ss,DX,(Bit16u)read_byte(BIOSMEM_SEG,BIOSMEM_CHAR_HEIGHT));
+ write_word(ss,CX,(Bit16u)read_byte(BIOSMEM_SEG,BIOSMEM_CHAR_HEIGHT));
 
  // Set Highest char row
  write_word(ss,DX,(Bit16u)read_byte(BIOSMEM_SEG,BIOSMEM_NB_ROWS));
@@ -1982,7 +2276,7 @@ static void memsetb(seg,offset,value,count)
   Bit16u value;
   Bit16u count;
 {
-#asm
+ASM_START
   push bp
   mov  bp, sp
 
@@ -2010,7 +2304,7 @@ memsetb_end:
     pop ax
 
   pop bp
-#endasm
+ASM_END
 }
 
 // --------------------------------------------------------------------------------------------
@@ -2020,7 +2314,7 @@ static void memsetw(seg,offset,value,count)
   Bit16u value;
   Bit16u count;
 {
-#asm
+ASM_START
   push bp
   mov  bp, sp
 
@@ -2048,7 +2342,7 @@ memsetw_end:
     pop ax
 
   pop bp
-#endasm
+ASM_END
 }
 
 // --------------------------------------------------------------------------------------------
@@ -2059,7 +2353,7 @@ static void memcpyb(dseg,doffset,sseg,soffset,count)
   Bit16u soffset;
   Bit16u count;
 {
-#asm
+ASM_START
   push bp
   mov  bp, sp
 
@@ -2094,7 +2388,7 @@ memcpyb_end:
     pop ax
 
   pop bp
-#endasm
+ASM_END
 }
 
 // --------------------------------------------------------------------------------------------
@@ -2105,7 +2399,7 @@ static void memcpyw(dseg,doffset,sseg,soffset,count)
   Bit16u soffset;
   Bit16u count;
 {
-#asm
+ASM_START
   push bp
   mov  bp, sp
 
@@ -2140,7 +2434,7 @@ memcpyw_end:
     pop ax
 
   pop bp
-#endasm
+ASM_END
 }
 
 /* =========================================================== */
@@ -2155,7 +2449,7 @@ read_byte(seg, offset)
   Bit16u seg;
   Bit16u offset;
 {
-#asm
+ASM_START
   push bp
   mov  bp, sp
 
@@ -2170,7 +2464,7 @@ read_byte(seg, offset)
     pop  bx
 
   pop  bp
-#endasm
+ASM_END
 }
 
 // --------------------------------------------------------------------------------------------
@@ -2179,7 +2473,7 @@ read_word(seg, offset)
   Bit16u seg;
   Bit16u offset;
 {
-#asm
+ASM_START
   push bp
   mov  bp, sp
 
@@ -2194,7 +2488,7 @@ read_word(seg, offset)
     pop  bx
 
   pop  bp
-#endasm
+ASM_END
 }
 
 // --------------------------------------------------------------------------------------------
@@ -2204,7 +2498,7 @@ write_byte(seg, offset, data)
   Bit16u offset;
   Bit8u  data;
 {
-#asm
+ASM_START
   push bp
   mov  bp, sp
 
@@ -2221,7 +2515,7 @@ write_byte(seg, offset, data)
     pop  ax
 
   pop  bp
-#endasm
+ASM_END
 }
 
 // --------------------------------------------------------------------------------------------
@@ -2231,7 +2525,7 @@ write_word(seg, offset, data)
   Bit16u offset;
   Bit16u data;
 {
-#asm
+ASM_START
   push bp
   mov  bp, sp
 
@@ -2248,7 +2542,7 @@ write_word(seg, offset, data)
     pop  ax
 
   pop  bp
-#endasm
+ASM_END
 }
 
 // --------------------------------------------------------------------------------------------
@@ -2256,7 +2550,7 @@ write_word(seg, offset, data)
 inb(port)
   Bit16u port;
 {
-#asm
+ASM_START
   push bp
   mov  bp, sp
 
@@ -2266,14 +2560,14 @@ inb(port)
     pop  dx
 
   pop  bp
-#endasm
+ASM_END
 }
 
   Bit16u
 inw(port)
   Bit16u port;
 {
-#asm
+ASM_START
   push bp
   mov  bp, sp
 
@@ -2283,7 +2577,7 @@ inw(port)
     pop  dx
 
   pop  bp
-#endasm
+ASM_END
 }
 
 // --------------------------------------------------------------------------------------------
@@ -2292,7 +2586,7 @@ outb(port, val)
   Bit16u port;
   Bit8u  val;
 {
-#asm
+ASM_START
   push bp
   mov  bp, sp
 
@@ -2305,7 +2599,7 @@ outb(port, val)
     pop  ax
 
   pop  bp
-#endasm
+ASM_END
 }
 
 // --------------------------------------------------------------------------------------------
@@ -2314,7 +2608,7 @@ outw(port, val)
   Bit16u port;
   Bit16u  val;
 {
-#asm
+ASM_START
   push bp
   mov  bp, sp
 
@@ -2327,14 +2621,14 @@ outw(port, val)
     pop  ax
 
   pop  bp
-#endasm
+ASM_END
 }
 
 Bit16u get_SS()
 {
-#asm
+ASM_START
   mov  ax, ss
-#endasm
+ASM_END
 }
 
 #ifdef DEBUG
@@ -2349,7 +2643,6 @@ void unknown()
 }
 #endif
 
-#ifdef DEBUG
 // --------------------------------------------------------------------------------------------
 void printf(s)
   Bit8u *s;
@@ -2385,9 +2678,9 @@ void printf(s)
         for (i=0; i<format_width; i++) {
           nibble = (arg >> (4 * digit)) & 0x000f;
           if (nibble <= 9)
-            outb(0xfff0, nibble + '0');
+            outb(0x0500, nibble + '0');
           else
-            outb(0xfff0, (nibble - 10) + 'A');
+            outb(0x0500, (nibble - 10) + 'A');
           digit--;
           }
         in_format = 0;
@@ -2397,12 +2690,11 @@ void printf(s)
       //  }
       }
     else {
-      outb(0xfff0, c);
+      outb(0x0500, c);
       }
     s ++;
     }
 }
-#endif
 
 #ifdef VBE
 #include "vbe.c"
@@ -2410,15 +2702,14 @@ void printf(s)
 
 // --------------------------------------------------------------------------------------------
 
-#asm 
+ASM_START 
 ;; DATA_SEG_DEFS_HERE
-#endasm
+ASM_END
 
-#asm
+ASM_START
 .ascii "vgabios ends here"
 .byte  0x00
 vgabios_end:
 .byte 0xCB
 ;; BLOCK_STRINGS_BEGIN
-#endasm
-
+ASM_END
